#include "preset.h"
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define INIT_PRESETS_CAPACITY 2

static char presets_path[PATH_MAX];

void get_presets_path(char *path) {
	get_xdg_path(path, "XDG_DATA_HOME", ".local/share", "tick/presets.ini");
}

void write_or_error(int fd, const void *buf) {
	size_t len = strlen(buf);
	ssize_t ret = write(fd, buf, len);
	if (ret == -1) {
		perror("Failed to write to presets");
	} else if ((size_t) ret < len) {
		fprintf(stderr, "Short write to presets");
	}
}

void add_to_presets(struct Presets *presets, struct Preset *preset) {
	if (presets->size == presets->capacity) {
		presets->capacity *= 2;
		presets->items = realloc(presets->items, sizeof(struct Preset) * presets->capacity);
	}
	presets->items[presets->size] = *preset;
	presets->size += 1;
}

void rename_preset(struct Preset *preset, char *name) {
	char tmp_path[PATH_MAX] = "tick-presets";
	int fd = mktemp_in_tmpdir(tmp_path);
	FILE *presets_file = fopen(presets_path, "r");

	char line[MAX_LINE_LEN];
	int found = 0;
	char *pos;
	while (fgets(line, MAX_LINE_LEN, presets_file) != NULL) {
		if (!found) {
			pos = strrchr(line, ']');
			if (pos != NULL) {
				*pos = '\0';
				if (strcmp(preset->name, line+1) == 0) {
					dprintf(fd, "[%s]\n", name);
					found = 1;
					continue;
				}
				*pos = ']';
			}
		}
		write_or_error(fd, line);
	}
	fclose(presets_file);
	close(fd);
	rename_file(tmp_path, presets_path);

	memcpy(preset->name, name, MAX_PRESET_NAME_LEN);
}

void edit_preset_settings(struct Preset *preset, int bpm, char *rhythm) {
	char tmp_path[PATH_MAX] = "tick-presets";
	int fd = mktemp_in_tmpdir(tmp_path);
	FILE *presets_file = fopen(presets_path, "r");

	char line[MAX_LINE_LEN];
	int found = 0;
	int in_section = 0;
	char *pos;
	while (fgets(line, MAX_LINE_LEN, presets_file) != NULL) {
		if (!in_section) {
			write_or_error(fd, line);
		} else if (line[0] == '[') {
			dprintf(fd, "bpm=%d\nrhythm=%s\n\n", bpm, rhythm);
			write_or_error(fd, line);
			in_section = 0;
		}
		if (!found) {
			pos = strrchr(line, ']');
			if (pos != NULL) {
				*pos = '\0';
				if (strcmp(preset->name, line+1) == 0) {
					found = 1;
					in_section = 1;
				}
			}
		}
	}
	if (in_section) {
		dprintf(fd, "bpm=%d\nrhythm=%s\n", bpm, rhythm);
	}
	fclose(presets_file);
	close(fd);
	rename_file(tmp_path, presets_path);

	preset->bpm = bpm;
	memcpy(preset->rhythm, rhythm, MAX_RHYTHM_LEN);
}

void create_preset(struct Presets *presets, char *name, int bpm, char *rhythm) {
	if (mkpath(presets_path) == -1) {
		perror("Failed to create necessary directories");
	}
	FILE *file = fopen(presets_path, "a");

	char str[MAX_LINE_LEN * 3 + 2]; // 2 from 2 '\n'
	sprintf(str, "\n[%s]\nbpm=%d\nrhythm=%s\n", name, bpm, rhythm);
	fputs(str, file);
	fclose(file);

	struct Preset preset = { .bpm = bpm };
	memcpy(preset.name, name, MAX_PRESET_NAME_LEN);
	memcpy(preset.rhythm, rhythm, MAX_RHYTHM_LEN);
	add_to_presets(presets, &preset);
}

void delete_preset(struct Presets *presets, int i) {
	char tmp_path[PATH_MAX] = "tick-presets";
	int fd = mktemp_in_tmpdir(tmp_path);
	FILE *presets_file = fopen(presets_path, "r");

	char line[MAX_LINE_LEN];
	int found = 0;
	int in_section = 0;
	char *pos;
	while (fgets(line, MAX_LINE_LEN, presets_file) != NULL) {
		if (!found) {
			pos = strrchr(line, ']');
			if (pos != NULL) {
				*pos = '\0';
				if (strcmp(presets->items[i].name, line+1) == 0) {
					found = 1;
					in_section = 1;
					continue;
				}
				*pos = ']';
			}
		}
		if (!in_section) {
			write_or_error(fd, line);
		} else if (line[0] == '[') {
			write_or_error(fd, line);
			in_section = 0;
		}
	}
	fclose(presets_file);
	close(fd);
	rename_file(tmp_path, presets_path);

	memmove(presets->items+i, presets->items+i+1, (presets->size-i-1)*sizeof(struct Preset));
	presets->size -= 1;
}

void process_presets_file(struct Presets *presets, FILE *file) {
	char line[MAX_LINE_LEN];
	char error[MAX_LINE_LEN+19]; // 19 from "unrecognized key ''"
	while (fgets(line, MAX_LINE_LEN, file) != NULL) {
		if (line[0] == '[') {
			char *pos = strrchr(line, ']');
			if (pos == NULL) {
				strcpy(error, "unmatched '['");
				goto invalid_config_exit;
			} else if (pos-1 == line) {
				strcpy(error, "preset name must be non-empty");
				goto invalid_config_exit;
			}
			struct Preset preset = { .bpm = 120, .rhythm = "" };
			memcpy(preset.name, line+1, pos-line-1);
			add_to_presets(presets, &preset);
		} else if (line[0] != '\n' && line[0] != '#') {
			char *pos = strchr(line, '=');
			if (pos == NULL) {
				strcpy(error, "expect key=value");
				goto invalid_config_exit;
			}
			*pos = '\0';
			pos += 1;
			if (strcmp(line, "bpm") == 0) {
				presets->items[presets->size-1].bpm = atoi(pos);
			} else if (strcmp(line, "rhythm") == 0) {
				// strnlen - 1 to remove \n if < max len, and leave room for \0 if > max len
				int len = strnlen(pos, MAX_RHYTHM_LEN) - 1;
				memcpy(presets->items[presets->size-1].rhythm, pos, len);
				presets->items[presets->size-1].rhythm[len] = '\0';
			} else {
				sprintf(error, "unrecognized key '%s'", line);
				goto invalid_config_exit;
			}
		}
	}
	return;

invalid_config_exit:
	fclose(file);
	free(presets->items);
	fprintf(stderr, "Invalid configuration: %s\n", error);
	exit(EXIT_FAILURE);
}

void get_presets(struct Presets *presets) {
	*presets = (struct Presets){
		.items = malloc(sizeof(struct Preset) * INIT_PRESETS_CAPACITY),
		.size = 0,
		.capacity = INIT_PRESETS_CAPACITY,
	};
	get_presets_path(presets_path);
	FILE *file = fopen(presets_path, "r");
	if (file != NULL) {
		process_presets_file(presets, file);
		fclose(file);
	}
	if (presets->size == 0) {
		presets->items[0] = (struct Preset){
			.name = "Default",
			.bpm = 120,
			.rhythm = ">...",
		};
	}
}
